@@ -26,7 +26,6 @@
 
 #include <stdlib.h>
 #include <stdio.h>
-#include <unistd.h>
 #include <string.h>
 #include <fcntl.h>
 #include <sys/types.h>
@@ -405,14 +404,7 @@ camera_abilities (CameraAbilitiesList *list)
 	for (i = 0; models[i].id_str; i++) {
 		memset (&a, 0, sizeof (a));
 
-		/* For now, flag EOS 20D as experimental. */
-		if ((UPLOAD_BOOL || (models[i].usb_capture_support == CAP_EXP)
-		     || models[i].model == CANON_CLASS_6) && 
-		    (models[i].usb_vendor && models[i].usb_product)) {
-			a.status = GP_DRIVER_STATUS_EXPERIMENTAL;
-		} else {
-			a.status = GP_DRIVER_STATUS_PRODUCTION;
-		}
+		a.status = GP_DRIVER_STATUS_PRODUCTION;
 
 		strcpy (a.model, models[i].id_str);
 		a.port = 0;
@@ -956,9 +948,7 @@ get_file_func (CameraFilesystem *fs, const char *folder, const char *filename,
 			break;
 #endif /* HAVE_LIBEXIF */
 		default:
-			/* this case should've been caught above anyway */
-			if (data)
-				free (data);
+			free (data);
 			data = NULL;
 			return (GP_ERROR_NOT_SUPPORTED);
 	}
@@ -1997,6 +1987,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Owner Name"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2010,6 +2001,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Capture Size Class"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 
 		i = 0;
@@ -2036,6 +2028,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("ISO Speed"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2065,6 +2058,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Shooting mode"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2073,7 +2067,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 			/* Map the menu option setting to the camera binary value */
 			i = 0;
 			while (shootingModeStateArray[i].label) {
-				if (strcmp (shootingModeStateArray[i].label, wvalue) == 0) {
+				if (strcmp (_(shootingModeStateArray[i].label), wvalue) == 0) {
 					shooting_mode = shootingModeStateArray[i].value;
 					break;
 				}
@@ -2094,6 +2088,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Shutter Speed"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2122,6 +2117,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Aperture"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2130,7 +2126,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 			/* Map the menu option setting to the camera binary value */
 			i = 0;
 			while (apertureStateArray[i].label) {
-				if (strcmp (apertureStateArray[i].label, wvalue) == 0) {
+				if (strcmp (_(apertureStateArray[i].label), wvalue) == 0) {
 					aperture = apertureStateArray[i].value;
 					break;
 				}
@@ -2150,6 +2146,8 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 	gp_widget_get_child_by_label (window, _("Exposure Compensation"), &w);
 	if (gp_widget_changed (w)) {
 		unsigned char expbias;
+
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2158,7 +2156,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 			/* Map the menu option setting to the camera binary value */
 			i = 0;
 			while (exposureBiasStateArray[i].label) {
-				if (strcmp (exposureBiasStateArray[i].label, wvalue) == 0) {
+				if (strcmp (_(exposureBiasStateArray[i].label), wvalue) == 0) {
 					expbias = exposureBiasStateArray[i].value;
 					break;
 				}
@@ -2178,6 +2176,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Image Format"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2186,7 +2185,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 			/* Map the menu option setting to the camera binary value */
 			i = 0;
 			while (imageFormatStateArray[i].label) {
-				if (strcmp (imageFormatStateArray[i].label, wvalue) == 0)
+				if (strcmp (_(imageFormatStateArray[i].label), wvalue) == 0)
 					break;
 				
 				i++;
@@ -2206,6 +2205,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Focus Mode"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2214,7 +2214,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 			/* Map the menu option setting to the camera binary value */
 			i = 0;
 			while (focusModeStateArray[i].label) {
-				if (strcmp (focusModeStateArray[i].label, wvalue) == 0) {
+				if (strcmp (_(focusModeStateArray[i].label), wvalue) == 0) {
 					focus_mode = focusModeStateArray[i].value;
 					break;
 				}
@@ -2235,6 +2235,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 	/* Beep */
 	gp_widget_get_child_by_label (window, _("Beep"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2268,6 +2269,8 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 		gp_widget_get_child_by_label (window, _("Zoom"), &w);
 		if (gp_widget_changed (w)) {
 			float zoom;
+
+	        	gp_widget_set_changed (w, 0);
 			gp_widget_get_value (w, &zoom);
 			if (!check_readiness (camera, context)) {
 				gp_context_status (context, _("Camera unavailable"));
@@ -2283,6 +2286,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 	/* Aperture */
 	gp_widget_get_child_by_label (window, _("Aperture"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2312,6 +2316,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 	/* Flash mode */
 	gp_widget_get_child_by_label (window, _("Flash Mode"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2341,6 +2346,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 
 	gp_widget_get_child_by_label (window, _("Synchronize camera date and time with PC"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &wvalue);
 		if (!check_readiness (camera, context)) {
 			gp_context_status (context, _("Camera unavailable"));
@@ -2356,6 +2362,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 	gp_widget_get_child_by_label (window, _("List all files"), &w);
 	if (gp_widget_changed (w)) {
 		/* XXXXX mark CameraFS as dirty */
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &val);
 		camera->pl->list_all_files = val;
 		sprintf (str, "%d", val);
@@ -2368,6 +2375,7 @@ camera_set_config (Camera *camera, CameraWidget *window, GPContext *context)
 #ifdef CANON_EXPERIMENTAL_UPLOAD
 	gp_widget_get_child_by_label (window, _("Keep filename on upload"), &w);
 	if (gp_widget_changed (w)) {
+	        gp_widget_set_changed (w, 0);
 		gp_widget_get_value (w, &camera->pl->upload_keep_filename);
 		GP_DEBUG ("New config value for \"Keep filename on upload\": %i",
 			  camera->pl->upload_keep_filename);
